(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.HtmlPageBuilder = {}));
})(this, (function (exports) { 'use strict';

  // ChatGPT
  function parseHTMLFragment(fragment) {
      // Create a new DOMParser
      const parser = new DOMParser();
      // Parse the HTML string into a DOM document
      const doc = parser.parseFromString(fragment, 'text/html');
      // Recursive function to parse DOM elements
      function parseElement(element) {
          const obj = {
              tag: element.tagName.toLowerCase(),
              attributes: [],
              children: [],
              cssClasses: element.classList.value.split(' ')
          };
          // Parse attributes
          for (const attr of element.attributes) {
              if (attr.name !== 'class') {
                  obj.attributes.push({ name: attr.name, value: attr.value });
              }
          }
          // Parse child nodes (both elements and text nodes)
          for (const child of element.childNodes) {
              if (child.nodeType === Node.ELEMENT_NODE) {
                  obj.children.push(parseElement(child)); // Recursive call for elements
              }
              else if (child.nodeType === Node.TEXT_NODE && child.textContent.trim()) {
                  obj.children.push({
                      tag: "#text",
                      attributes: [{ name: 'text', value: child.textContent.trim() }]
                  });
              }
          }
          return obj;
      }
      // Start parsing from the body of the document
      const result = [];
      for (const child of Array.from(doc.body.children)) {
          result.push(parseElement(child));
      }
      return result;
  }
  function kebabToCamelCase(str) {
      return str.replace(/-([a-z])/g, function (match, letter) {
          return letter.toUpperCase();
      });
  }
  function camelToKebabCase(str) {
      return str.replace(/([A-Z])/g, '-$1').toLowerCase();
  }
  function hexToRgb(hex) {
      // Remove the '#' if it's included in the hex value
      hex = hex.replace(/^#/, '');
      // If the hex is shorthand (e.g. #03F), expand it to full form (e.g. #0033FF)
      if (hex.length === 3) {
          hex = hex.split('').map(char => char + char).join('');
      }
      // Parse the r, g, b values from the hex string
      const bigint = parseInt(hex, 16);
      const r = (bigint >> 16) & 255;
      const g = (bigint >> 8) & 255;
      const b = bigint & 255;
      return `rgb(${r}, ${g}, ${b})`;
  }
  function isHexColor(str) {
      // Regular expression to match valid hex color codes (#RRGGBB or #RGB)
      const hexRegex = /^#([A-Fa-f0-9]{3}){1,2}$/;
      return hexRegex.test(str);
  }
  function isCssPropNameValid(element, propertyName) {
      const propToCheck = kebabToCamelCase(propertyName);
      const hasProp = element.style.hasOwnProperty(propToCheck);
      return hasProp;
  }
  function isCssPropValueValid(propertyName, propertyValue) {
      const propToCheck = camelToKebabCase(propertyName);
      const propValueToCheck = propertyValue.split(' ').map(item => isHexColor(item) ? hexToRgb(item) : item).join(' ');
      if (CSS.supports(propToCheck, propValueToCheck)) {
          return true;
      }
      // Optionally check for vendor prefixes
      const prefixes = ['-webkit-', '-moz-', '-ms-', '-o-'];
      return prefixes.some(prefix => CSS.supports(prefix + propToCheck, propertyValue));
  }

  exports.SystemAttributes = void 0;
  (function (SystemAttributes) {
      SystemAttributes["nodeKey"] = "data-key";
  })(exports.SystemAttributes || (exports.SystemAttributes = {}));

  class EventEmitter {
      constructor() {
          this.listenersMap = new Map();
      }
      on(eventNames, handler, options) {
          const eventNameList = eventNames.split(' ');
          eventNameList.forEach(eventName => {
              const listeners = this.getEventListeners(eventName);
              listeners.push({ handler, autoUnlistenWhen: options?.autoUnlistenWhen });
              this.listenersMap.set(eventName, listeners);
          });
          return () => this.unlisten(eventNames, handler);
      }
      once(eventName, handler) {
          const listeners = this.getEventListeners(eventName);
          listeners.push({ handler, once: true });
          return () => this.unlisten(eventName, handler);
      }
      emit(eventName, payload) {
          const listeners = this.getEventListeners(eventName);
          listeners.forEach(listener => listener.handler(payload));
          this.listenersMap.set(eventName, listeners.filter(listener => !listener.once || (listener.autoUnlistenWhen && listener.autoUnlistenWhen())));
      }
      getEventListeners(eventName) {
          return this.listenersMap.get(eventName) ?? [];
      }
      unlisten(eventNames, handler) {
          const eventNameList = eventNames.split(' ');
          eventNameList.forEach(eventName => {
              this.listenersMap.set(eventName, this.getEventListeners(eventName).filter(l => l.handler !== handler));
          });
      }
  }

  var NodeType;
  (function (NodeType) {
      NodeType[NodeType["text"] = 0] = "text";
      NodeType[NodeType["element"] = 1] = "element";
      NodeType[NodeType["custom"] = 2] = "custom";
  })(NodeType || (NodeType = {}));

  /**
   * Convert array of 16 byte values to UUID string format of the form:
   * XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
   */
  var byteToHex = [];
  for (var i = 0; i < 256; ++i) {
    byteToHex.push((i + 0x100).toString(16).slice(1));
  }
  function unsafeStringify(arr, offset = 0) {
    // Note: Be careful editing this code!  It's been tuned for performance
    // and works in ways you may not expect. See https://github.com/uuidjs/uuid/pull/434
    //
    // Note to future-self: No, you can't remove the `toLowerCase()` call.
    // REF: https://github.com/uuidjs/uuid/pull/677#issuecomment-1757351351
    return (byteToHex[arr[offset + 0]] + byteToHex[arr[offset + 1]] + byteToHex[arr[offset + 2]] + byteToHex[arr[offset + 3]] + '-' + byteToHex[arr[offset + 4]] + byteToHex[arr[offset + 5]] + '-' + byteToHex[arr[offset + 6]] + byteToHex[arr[offset + 7]] + '-' + byteToHex[arr[offset + 8]] + byteToHex[arr[offset + 9]] + '-' + byteToHex[arr[offset + 10]] + byteToHex[arr[offset + 11]] + byteToHex[arr[offset + 12]] + byteToHex[arr[offset + 13]] + byteToHex[arr[offset + 14]] + byteToHex[arr[offset + 15]]).toLowerCase();
  }

  // Unique ID creation requires a high quality random # generator. In the browser we therefore
  // require the crypto API and do not support built-in fallback to lower quality random number
  // generators (like Math.random()).

  var getRandomValues;
  var rnds8 = new Uint8Array(16);
  function rng() {
    // lazy load so that environments that need to polyfill have a chance to do so
    if (!getRandomValues) {
      // getRandomValues needs to be invoked in a context where "this" is a Crypto implementation.
      getRandomValues = typeof crypto !== 'undefined' && crypto.getRandomValues && crypto.getRandomValues.bind(crypto);
      if (!getRandomValues) {
        throw new Error('crypto.getRandomValues() not supported. See https://github.com/uuidjs/uuid#getrandomvalues-not-supported');
      }
    }
    return getRandomValues(rnds8);
  }

  var randomUUID = typeof crypto !== 'undefined' && crypto.randomUUID && crypto.randomUUID.bind(crypto);
  var native = {
    randomUUID
  };

  function v4(options, buf, offset) {
    if (native.randomUUID && !buf && !options) {
      return native.randomUUID();
    }
    options = options || {};
    var rnds = options.random || (options.rng || rng)();

    // Per 4.4, set bits for version and `clock_seq_hi_and_reserved`
    rnds[6] = rnds[6] & 0x0f | 0x40;
    rnds[8] = rnds[8] & 0x3f | 0x80;
    return unsafeStringify(rnds);
  }

  class PageBaseNode extends EventEmitter {
      constructor() {
          super();
      }
      createNode() {
          this.key = v4();
          this.schema.key = this.key;
      }
      findElement(predicate) {
          return predicate(this) ? this : undefined;
      }
      ;
      destroy() {
          this.parentElement?.removeChild(this);
      }
      getSchema() {
          return this.schema;
      }
      removeChild(child) {
          this.children = this.children?.filter(el => el !== child);
          this.nativeElement.removeChild(child.hostElement || child.nativeElement);
          this.page.schemaChanged();
      }
      changeElementIndex(newIndex) {
          if (!this.parentElement) {
              return;
          }
          if (newIndex < 0 || newIndex > Number(this.parentElement?.children?.length)) {
              console.error('Invalid indices provided');
              return;
          }
          const parent = this.parentElement;
          const currentIndex = parent.children.findIndex(el => el === this);
          const targetPageNode = parent.children[newIndex];
          const targetElement = targetPageNode.hostElement;
          const currentElement = this.hostElement;
          parent.nativeElement.removeChild(currentElement);
          if (currentIndex > newIndex) {
              parent.nativeElement.insertBefore(currentElement, targetElement);
          }
          else {
              const nextSibling = targetElement.nextElementSibling;
              if (!nextSibling) {
                  parent.nativeElement.appendChild(currentElement);
              }
              else {
                  parent.nativeElement.insertBefore(currentElement, nextSibling);
              }
          }
          parent.children = parent.children.filter(it => it !== this);
          parent.children.splice(newIndex, 0, this);
          this.page.schemaChanged();
      }
  }

  class ElementAttributes {
      get value() {
          return this.pageElement.schema.attributes || [];
      }
      get isNodeWithAttributes() {
          return this.pageElement.nativeElement.nodeType !== Node.TEXT_NODE;
      }
      get hostElement() {
          return this.pageElement.hostElement;
      }
      constructor(pageElement) {
          this.pageElement = pageElement;
          this.attributesMap = new Map();
      }
      add(...attributes) {
          this.setAttributes(attributes);
          this.updateSchemaAttributes();
          this.pageElement.emit('view:change');
      }
      remove(...names) {
          this.removeAttributes(names);
          this.updateSchemaAttributes();
          this.pageElement.emit('view:change');
      }
      /**
       * @NotImplemented
       * @param name
       * @param force
       */
      toggle(name, force) {
          if (this.isNodeWithAttributes) {
              this.toggleAttribute(name, force);
          }
          this.pageElement.emit('view:change');
      }
      replace(name, attribute) {
          this.removeAttributes([name]);
          this.setAttributes([attribute]);
          this.updateSchemaAttributes();
          this.pageElement.emit('view:change');
      }
      init() {
          this.setAttributes(this.pageElement.schema.attributes);
      }
      updateSchemaAttributes() {
          this.pageElement.schema.attributes = Array.from(this.attributesMap).map(([name, value]) => ({ name, value }));
      }
      setAttributes(attrs) {
          attrs.forEach(attr => {
              if (this.isNodeWithAttributes) {
                  this.hostElement.setAttribute(attr.name, attr.value);
              }
              if (Object.values(exports.SystemAttributes).indexOf(attr.name) < 0) {
                  this.attributesMap.set(attr.name, attr.value);
              }
          });
      }
      removeAttributes(names) {
          names.forEach(name => {
              if (this.isNodeWithAttributes) {
                  this.hostElement.removeAttribute(name);
              }
              this.attributesMap.delete(name);
          });
      }
      toggleAttribute(name, force) {
          this.hostElement.toggleAttribute(name, force);
      }
  }

  class PageTextNode extends PageBaseNode {
      get hostElement() {
          return this.wrapper || this.nativeElement;
      }
      constructor(schema, page, parentElement) {
          super();
          this.parentElement = parentElement;
          this.nodeType = NodeType.text;
          this.page = page;
          this.schema = schema;
          this.attributes = new ElementAttributes(this);
          this.createTextNode();
      }
      wrap(wrapperTag) {
          this.parentElement?.nativeElement.removeChild(this.nativeElement);
          this.wrapper = this.page.documentRef.createElement(wrapperTag);
          this.wrapper.setAttribute('data-key', this.key);
          this.wrapper.appendChild(this.nativeElement);
          this.parentElement?.nativeElement.appendChild(this.hostElement);
          this.wrapper.tabIndex = 1;
          return this.wrapper;
      }
      getValue() {
          return this.schema.attributes?.find(attr => attr.name === 'text')?.value ?? '';
      }
      createTextNode() {
          super.createNode();
          this.nativeElement = this.page.documentRef.createTextNode(this.getValue());
          this.parentElement?.nativeElement.appendChild(this.nativeElement);
          this.listenSelfEvents();
          this.page.nodeCreated(this);
      }
      listenSelfEvents() {
          this.on('view:change', () => this.onViewChange());
      }
      onViewChange() {
          const value = this.getValue();
          if (value !== this.nativeElement.textContent) {
              this.nativeElement.textContent = value;
          }
      }
  }

  class ElementClassList {
      get value() {
          return this.hostElement.classList.value ?? '';
      }
      get hostElement() {
          return this.pageElement.hostElement;
      }
      constructor(pageElement) {
          this.pageElement = pageElement;
      }
      getClassNames() {
          return this.value.split(' ')
              .filter(className => {
              const includeSystemClasses = className.includes(this.pageElement.key);
              if (includeSystemClasses) {
                  return false;
              }
              const isIgnored = this.pageElement.page.getOptions().ignoreCssClasses?.includes(className);
              return !isIgnored;
          });
      }
      init() {
          this.hostElement.classList.add(...(this.pageElement.schema.cssClasses || []));
      }
      add(...classNames) {
          this.hostElement.classList.add(...classNames);
          this.updateSchemaClassList();
          this.pageElement.emit('view:change');
      }
      remove(...classNames) {
          this.hostElement.classList.remove(...classNames);
          this.updateSchemaClassList();
          this.pageElement.emit('view:change');
      }
      toggle(className, force) {
          this.hostElement.classList.toggle(className, force);
          this.updateSchemaClassList();
          this.pageElement.emit('view:change');
      }
      updateSchemaClassList() {
          this.pageElement.schema.cssClasses = this.getClassNames();
      }
  }

  var StyleType;
  (function (StyleType) {
      StyleType["border"] = "border";
      StyleType["layout"] = "layout";
      StyleType["text"] = "text";
      StyleType["boxShadow"] = "boxShadow";
      StyleType["backgroundBlur"] = "backgroundBlur";
      StyleType["layerBlur"] = "layerBlur";
      StyleType["custom"] = "custom";
      StyleType["size"] = "size";
  })(StyleType || (StyleType = {}));

  class StyleRule {
      constructor() {
          this.rule = {};
          this.detachedProps = new Set();
      }
      getAllProps() {
          return Object.keys(this.rule).filter(key => !!this.rule[key]).reduce((r, key) => {
              r[key] = this.rule[key];
              return r;
          }, {});
      }
      setProp(key, value) {
          this.rule[key] = value;
      }
      getProp(key) {
          return this.rule[key];
      }
      getPropOrPropEdges(key, options) {
          return this.rule[key] ?? this.getPropEdges(key, options);
      }
      getPropEdges(key, options) {
          const edges = options?.edgeKeys ?? ['left', 'top', 'right', 'bottom'];
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          const edgesRule = edges.reduce((r, edge) => {
              const edgePropKey = edgePropKeyFactory(edge);
              if (this.rule[edgePropKey]) {
                  r[edge] = this.rule[edgePropKey];
              }
              return r;
          }, {});
          return Object.keys(edgesRule).length ? edgesRule : undefined;
      }
      setPropOrPropEdges(key, value, options) {
          if (typeof value === 'string') {
              this.deletePropEdges(key, options);
              this.rule[key] = value;
              return;
          }
          this.deleteProp(key);
          this.setPropEdges(key, value, options);
      }
      setPropEdges(key, value, options) {
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          Object.entries(value).forEach(([edgeKey, value]) => {
              const edgePropKey = edgePropKeyFactory(edgeKey);
              this.rule[edgePropKey] = value;
          });
      }
      attachProp(key) {
          this.detachedProps.delete(key);
      }
      detachProp(key) {
          this.detachedProps.add(key);
      }
      detached(key) {
          return this.detachedProps.has(key);
      }
      init(rule) {
          this.rule = rule;
      }
      deleteProp(key) {
          delete this.rule[key];
      }
      deletePropEdges(key, options) {
          const edges = options?.edgeKeys ?? ['left', 'top', 'right', 'bottom'];
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          edges.forEach(edgeKey => {
              const edgePropKey = edgePropKeyFactory(edgeKey);
              delete this.rule[edgePropKey];
          });
      }
      entries() {
          return Object.entries(this.rule);
      }
      generateRuleText() {
          return this.entries()
              .filter(([key, value]) => !this.detachedProps.has(key) && !!value)
              .reduce((result, [key, value]) => {
              const ruleText = `${key}: ${value};`;
              result += ruleText;
              return result;
          }, '');
      }
  }

  class StyleBase {
      constructor() {
          this.rule = new StyleRule();
          this.ruleIndex = -1;
      }
      setStyleRule(ruleValue) {
          this.rule.init(ruleValue);
          this.applyStyleRule();
      }
      applyStyles() {
          this.applyStyleRule();
          this.element.page.schemaChanged();
      }
      tryToApplyStyleRule(apply = true) {
          if (apply) {
              this.applyStyles();
          }
      }
      applyStyleRule() {
          const ruleText = this.rule.generateRuleText();
          if (!ruleText?.length && this.ruleIndex >= 0) {
              this.hardDeleteRule();
              this.updateNodeSchema();
              this.element.emit('view:change');
              return;
          }
          if (ruleText === this.ruleText) {
              return;
          }
          this.ruleText = ruleText;
          if (this.ruleIndex >= 0) {
              this.updateRule();
          }
          else {
              this.insertRule();
          }
          if (!this.element.nativeElement.classList.contains(this.selector)) {
              this.element.nativeElement.classList.add(this.selector);
          }
          this.updateNodeSchema();
          this.element.emit('view:change');
      }
      insertRule() {
          this.ruleIndex = this.styleSheet.insertRule(`.${this.selector}{${this.ruleText}}`, this.ruleIndex >= 0 ? this.ruleIndex : this.styleSheet.cssRules.length);
      }
      updateRule() {
          this.deleteRule();
          this.insertRule();
      }
      deleteRule() {
          this.styleSheet.deleteRule(this.ruleIndex);
      }
      hardDeleteRule() {
          if (!this.styleSheet.cssRules.length) {
              console.warn('No rules were applied, nothing to delete');
              return;
          }
          this.deleteRule();
          this.ruleText = '';
          this.ruleIndex = -1;
      }
      updateNodeSchema() {
          if (!this.ruleText?.length) {
              delete this.element.schema.styles?.[this.styleType];
          }
          else {
              this.element.schema.styles = {
                  ...this.element.schema.styles,
                  [this.styleType]: this.rule.getAllProps()
              };
          }
      }
  }

  class BorderStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.border;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setWidth(width, apply = true) {
          this.setInitialValuesOnChange('width');
          if (typeof width !== 'string') {
              this.rule.setProp('border-width', '0px');
              this.rule.setPropEdges('border-width', width, {
                  getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
              });
          }
          else {
              this.rule.setProp('border-width', width);
              this.rule.deletePropEdges('border-width', {
                  getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
              });
          }
          this.tryToApplyStyleRule(apply);
      }
      getWidth() {
          return this.rule.getPropEdges('border-width', {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
          }) ?? this.rule.getProp('border-width');
      }
      setStyleType(styleType, apply = true) {
          this.setInitialValuesOnChange('styleType');
          this.rule.setPropOrPropEdges('border-style', styleType, {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-style`
          });
          this.tryToApplyStyleRule(apply);
      }
      getStyleType() {
          return this.rule.getPropOrPropEdges('border-style');
      }
      setColor(color, apply = true) {
          this.setInitialValuesOnChange('color');
          this.rule.setPropOrPropEdges('border-color', color, {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-color`
          });
          this.tryToApplyStyleRule(apply);
      }
      getColor() {
          return this.rule.getPropOrPropEdges('border-color', {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-color`,
          });
      }
      setBorderRadius(radius, apply = true) {
          this.setInitialValuesOnChange('border-radius');
          this.rule.setPropOrPropEdges('border-radius', radius, {
              edgeKeys: ['topRight', 'topLeft', 'bottomRight', 'bottomLeft'],
              getEdgePropKey: (edgeKey) => `border-${camelToKebabCase(edgeKey)}-radius`
          });
          this.tryToApplyStyleRule(apply);
      }
      getBorderRadius() {
          return this.rule.getPropOrPropEdges('border-radius', {
              edgeKeys: ['topRight', 'topLeft', 'bottomRight', 'bottomLeft'],
              getEdgePropKey: (edgeKey) => `border-${camelToKebabCase(edgeKey)}-radius`
          });
      }
      setInitialValuesOnChange(changedProp) {
          if (changedProp !== 'width' && this.getWidth() === undefined) {
              this.rule.setProp('border-width', '1px');
          }
          if (changedProp !== 'styleType' && this.getStyleType() === undefined) {
              this.rule.setProp('border-style', 'solid');
          }
          if (changedProp !== 'color' && this.getColor() === undefined) {
              this.rule.setProp('border-color', '#000000');
          }
      }
  }

  class BoxShadowStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.boxShadow;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      has(effect) {
          const str = this.effectToString(effect);
          return !!this.rule.getProp('box-shadow')?.includes(str);
      }
      set(effect, apply = true) {
          const shadowProp = this.effectToString(effect);
          const boxShadowValue = this.rule.getProp('box-shadow') ?? '';
          const updatedBoxShadowValue = `${boxShadowValue}${boxShadowValue?.length ? ', ' : ''}${shadowProp}`;
          this.rule.setProp('box-shadow', updatedBoxShadowValue);
          this.tryToApplyStyleRule(apply);
      }
      getEffects() {
          return this.rule.getProp('box-shadow')?.split(', ').map((prop) => this.stringToEffect(prop)) || [];
      }
      delete(effect, apply = true) {
          if (!this.rule.getProp('box-shadow')?.length) {
              return;
          }
          const shadowProp = this.effectToString(effect);
          const boxShadowProps = this.rule.getProp('box-shadow').split(', ');
          this.rule.setProp('box-shadow', boxShadowProps.filter((prop) => shadowProp !== prop).join(', '));
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(effect) {
          return isCssPropValueValid('box-shadow', this.effectToString(effect));
      }
      effectToString(shadow) {
          return `${shadow.inner ? 'inset ' : ''}${shadow.x} ${shadow.y} ${shadow.blur} ${shadow.spread} ${shadow.color}`;
      }
      stringToEffect(str) {
          const items = str.split(' ');
          const inner = items[0] === 'inset';
          const [x, y, blur, spread, color] = items.filter(item => item !== 'inset');
          return { inner, x, y, blur, spread, color };
      }
  }

  class CustomStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.custom;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      get(property) {
          return this.rule.getProp(property);
      }
      getAll() {
          return this.rule.getAllProps();
      }
      append(property, value, apply = true) {
          this.rule.setProp(property, value);
          this.tryToApplyStyleRule(apply);
      }
      delete(property, apply = true) {
          this.rule.deleteProp(property);
          this.tryToApplyStyleRule(apply);
      }
      attach(key) {
          this.rule.attachProp(key);
          super.applyStyles();
      }
      detach(key) {
          this.rule.detachProp(key);
          super.applyStyles();
      }
      detached(key) {
          return this.rule.detached(key);
      }
  }

  class LayoutStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.layout;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setPosition({ direction, x, y, gap }, apply = true) {
          this.rule.setProp('display', 'flex');
          this.rule.setProp('flex-direction', direction);
          this.rule.setProp('gap', gap ?? '');
          if (direction === 'row') {
              this.rule.setProp('justify-content', x === 'space-between' || x === 'center' ? x : `flex-${x}`);
              this.rule.setProp('align-items', y === 'center' ? y : `flex-${y}`);
          }
          else if (direction === 'column') {
              this.rule.setProp('justify-content', y === 'center' || y === 'space-between' ? y : `flex-${y}`);
              this.rule.setProp('align-items', x === 'center' ? x : `flex-${x}`);
          }
          this.tryToApplyStyleRule(apply);
      }
      getPosition() {
          if (this.rule.getProp('display') !== 'flex') {
              return undefined;
          }
          const p = {};
          p.direction = this.rule.getProp('flex-direction');
          let xRule, yRule;
          if (p.direction === 'row') {
              xRule = this.rule.getProp('justify-content');
              yRule = this.rule.getProp('align-items');
          }
          else {
              xRule = this.rule.getProp('align-items');
              yRule = this.rule.getProp('justify-content');
          }
          p.x = xRule === 'space-between' || xRule === 'center' ? xRule : xRule.split('-')[1];
          p.y = yRule === 'space-between' || yRule === 'center' ? yRule : yRule.split('-')[1];
          p.gap = this.rule.getProp('gap');
          return p;
      }
      deletePosition(apply = true) {
          this.rule.deleteProp('display');
          this.rule.deleteProp('flex-direction');
          this.rule.deleteProp('justify-content');
          this.rule.deleteProp('align-items');
          this.rule.deleteProp('gap');
          this.tryToApplyStyleRule(apply);
      }
      setPadding(padding, apply = true) {
          this.rule.setPropOrPropEdges('padding', padding);
          this.tryToApplyStyleRule(apply);
      }
      getPadding() {
          return this.rule.getPropOrPropEdges('padding');
      }
      deletePadding(apply = true) {
          this.rule.deleteProp('padding');
          this.rule.deletePropEdges('padding');
          this.tryToApplyStyleRule(apply);
      }
      setMargin(margin, apply = true) {
          this.rule.setPropOrPropEdges('margin', margin);
          this.tryToApplyStyleRule(apply);
      }
      getMargin() {
          return this.rule.getPropOrPropEdges('margin');
      }
      deleteMargin(apply = true) {
          this.rule.deleteProp('margin');
          this.tryToApplyStyleRule(apply);
      }
      setBackgroundColor(color, apply = true) {
          this.rule.setProp('background-color', color);
          this.tryToApplyStyleRule(apply);
      }
      getBackgroundColor() {
          return this.rule.getProp('background-color');
      }
      deleteBackgroundColor(apply = true) {
          this.rule.deleteProp('background-color');
          this.tryToApplyStyleRule(apply);
      }
  }

  class TextStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.text;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
          this.setDefaults();
      }
      getColor() {
          return this.rule.getProp('color');
      }
      setColor(color, apply = true) {
          this.rule.setProp('color', color);
          this.tryToApplyStyleRule(apply);
      }
      getLineHeight() {
          return this.rule.getProp('line-height');
      }
      setLineHeight(lineHeight, apply = true) {
          this.rule.setProp('line-height', lineHeight);
          this.tryToApplyStyleRule(apply);
      }
      getAlignment() {
          return this.rule.getProp('text-align');
      }
      setAlignment(alignment, apply = true) {
          this.rule.setProp('text-align', alignment);
          this.tryToApplyStyleRule(apply);
      }
      getFontSize() {
          return this.rule.getProp('font-size');
      }
      setFontSize(fontSize, apply = true) {
          this.rule.setProp('font-size', fontSize);
          this.tryToApplyStyleRule(apply);
      }
      getDecoration() {
          return this.rule.getProp('text-decoration');
      }
      setDecoration(decoration, apply = true) {
          this.rule.setProp('text-decoration', decoration);
          this.tryToApplyStyleRule(apply);
      }
      setDefaults() {
          this.rule.setProp('color', '#000000');
          this.rule.setProp('line-height', 'auto');
          this.rule.setProp('text-align', 'left');
      }
  }

  class BackgroundBlurStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.backgroundBlur;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      set(blur, apply = true) {
          this.rule.setProp('backdrop-filter', `blur(${blur})`);
          this.tryToApplyStyleRule(apply);
      }
      get() {
          return this.getEffectValue();
      }
      delete(apply = true) {
          this.rule.deleteProp('backdrop-filter');
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(value) {
          return isCssPropValueValid('backdrop-filter', `blur(${value})`);
      }
      getEffectValue() {
          if (!this.rule.getProp('backdrop-filter')) {
              return '';
          }
          const match = this.rule.getProp('backdrop-filter').match(/\(([^)]+)\)/);
          return match[1];
      }
  }

  class LayerBlurStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.layerBlur;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      set(blur, apply = true) {
          this.rule.setProp('filter', `blur(${blur})`);
          this.tryToApplyStyleRule(apply);
      }
      get() {
          return this.getEffectValue();
      }
      delete(apply = true) {
          this.rule.deleteProp('filter');
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(value) {
          return isCssPropValueValid('filter', `blur(${value})`);
      }
      getEffectValue() {
          if (!this.rule.getProp('filter')) {
              return '';
          }
          const match = this.rule.getProp('filter').match(/\(([^)]+)\)/);
          return match[1];
      }
  }

  class SizeStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = StyleType.size;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setHeight(height, apply = true) {
          this.rule.setProp('height', height);
          this.tryToApplyStyleRule(apply);
      }
      getHeight() {
          return this.rule.getProp('height');
      }
      deleteHeight(apply = true) {
          this.rule.deleteProp('height');
          this.tryToApplyStyleRule(apply);
      }
      setWidth(width, apply = true) {
          this.rule.setProp('width', width);
          this.tryToApplyStyleRule(apply);
      }
      getWidth() {
          return this.rule.getProp('width');
      }
      deleteWidth(apply = true) {
          this.rule.deleteProp('width');
          this.tryToApplyStyleRule(apply);
      }
  }

  function styleFactory(element, key) {
      switch (key) {
          case StyleType.border:
              return new BorderStyle(element);
          case StyleType.layout:
              return new LayoutStyle(element);
          case StyleType.text:
              return new TextStyle(element);
          case StyleType.boxShadow:
              return new BoxShadowStyle(element);
          case StyleType.backgroundBlur:
              return new BackgroundBlurStyle(element);
          case StyleType.layerBlur:
              return new LayerBlurStyle(element);
          case StyleType.size:
              return new SizeStyle(element);
          case StyleType.custom:
              return new CustomStyle(element);
          default:
              return undefined;
      }
  }

  class ElementStyles {
      constructor(pageElement) {
          this.pageElement = pageElement;
          this.stylesMap = new Map();
      }
      init() {
          Object.values(StyleType).forEach(type => {
              const style = styleFactory(this.pageElement, type);
              const schemaStyles = this.pageElement.schema.styles?.[type];
              this.stylesMap.set(type, style);
              if (schemaStyles) {
                  style.setStyleRule(schemaStyles);
              }
          });
      }
      get(type) {
          return this.stylesMap.get(type);
      }
  }

  class PageElementNode extends PageBaseNode {
      get childElements() {
          return this.children.filter(ch => ch instanceof PageElementNode);
      }
      constructor(schema, page, parentElement) {
          super();
          this.nodeType = NodeType.element;
          this.children = [];
          this.page = page;
          this.parentElement = parentElement;
          this.schema = schema;
          this.classList = new ElementClassList(this);
          this.styles = new ElementStyles(this);
          this.attributes = new ElementAttributes(this);
          this.createElementNode();
      }
      createElementNode() {
          super.createNode();
          const elRef = this.page.documentRef.createElement(this.schema.tag);
          this.nativeElement = elRef;
          this.hostElement = this.nativeElement;
          this.parentElement?.nativeElement?.appendChild(this.nativeElement);
          if (this.schema.children?.length) {
              this.children = this.schema.children.map(childSchema => this.insertChild(childSchema, false));
          }
          this.attributes.add({ name: exports.SystemAttributes.nodeKey, value: this.key });
          this.nativeElement.setAttribute(exports.SystemAttributes.nodeKey, this.key);
          this.classList.init();
          this.styles.init();
          this.attributes.init();
          this.nativeElement.tabIndex = 1;
          this.page.nodeCreated(this);
      }
      getSchema() {
          this.schema.children = this.children?.map(ch => ch.getSchema());
          return this.schema;
      }
      findElement(predicate) {
          const thisSuits = super.findElement(predicate);
          if (thisSuits) {
              return this;
          }
          const found = this.children?.map(ch => ch.findElement(predicate)).filter(Boolean)[0];
          if (found) {
              return found;
          }
          return undefined;
      }
      destroy() {
          super.destroy();
          this.page.emit('node:destroy', this);
          this.emit('destroy');
      }
      insertChild(schema, emitSchemaChanged = true) {
          const node = nodeFactory(schema, this.page, this);
          this.children = [...(this.children || []), node];
          this.schema.children = [...(this.schema.children || []), node.schema];
          if (emitSchemaChanged) {
              this.page.schemaChanged();
          }
          return node;
      }
      insertHTMLFragment(fragment) {
          const schemas = parseHTMLFragment(fragment);
          const children = schemas.map(schema => this.insertChild(schema, false));
          this.page.schemaChanged();
          return children;
      }
  }

  // import { cloneDeep } from "lodash";
  class PageCustomNode extends PageBaseNode {
      constructor(schema, page, parentElement) {
          super();
          this.nodeType = NodeType.custom;
          this.page = page;
          this.parentElement = parentElement;
          this.schema = schema;
          this.attributes = new ElementAttributes(this);
          this.createCustomNode();
      }
      createCustomNode() {
          const nodeTag = this.schema.tag.split('_')[1];
          if (!nodeTag) {
              console.warn('Page custom node tag was not specified');
              return;
          }
          super.createNode();
          const elRef = this.page.customElementFactory(this);
          this.nativeElement = elRef;
          this.hostElement = elRef;
          this.parentElement?.nativeElement?.appendChild(this.nativeElement);
          this.attributes.add({ name: exports.SystemAttributes.nodeKey, value: this.key });
          this.nativeElement.setAttribute(exports.SystemAttributes.nodeKey, this.key);
          this.nativeElement.tabIndex = 1;
          this.nativeElement.style.display = 'block';
          this.page.nodeCreated(this);
      }
      updateSchema(schema) {
          this.schema = schema;
          this.page.schemaChanged();
      }
  }

  function nodeFactory(schema, page, parentElement) {
      const tag = schema.tag.split('_')[0];
      switch (tag) {
          case 'custom':
              return new PageCustomNode(schema, page, parentElement);
          case '#text':
              return new PageTextNode(schema, page, parentElement);
          default:
              return new PageElementNode(schema, page, parentElement);
      }
  }

  class Page extends EventEmitter {
      get documentRef() {
          return this.iframe.contentDocument || this.iframe.contentWindow?.document;
      }
      get windowRef() {
          return this.iframe.contentWindow;
      }
      constructor(container, title, schema, options, rootDocumentRef) {
          super();
          this.container = container;
          this.title = title;
          this.schema = schema;
          this.options = options;
          this.rootDocumentRef = rootDocumentRef;
          this.libraries = [];
          this.registerIframe();
          this.iframe.addEventListener('load', () => {
              this.registerStyleSheet();
              this.emit('ready');
              this.linkLibrary({ name: 'custom-styles', href: './lib-styles.css', rel: 'stylesheet' });
          });
      }
      getRootElement() {
          return this.rootElement;
      }
      getStyleSheet() {
          return this.styleSheet;
      }
      getLinkedLibraries() {
          return this.libraries.map(lib => ({
              ...Array.from(lib.attributes).reduce((r, attr) => (r[attr.name] = attr.value) && r, {}),
              name: lib.id
          }));
      }
      getSchema() {
          return this.rootElement.getSchema();
      }
      schemaChanged() {
          this.emit('schema:change', this.getSchema());
      }
      getOptions() {
          return this.options;
      }
      buildElements(container) {
          this.rootElement = nodeFactory(this.schema, this);
          container.appendChild(this.rootElement.nativeElement);
      }
      customElementFactory(element) {
          if (!this.options.customElementFactory) {
              console.error('Custom element factory was not provided');
              return this.getStubElement();
          }
          return this.options.customElementFactory(element);
      }
      nodeCreated(node) {
          this.emit('node:create', node);
      }
      findElement(predicate) {
          return this.rootElement.findElement(predicate);
      }
      linkLibrary({ name, href, rel, ...rest }) {
          const link = this.documentRef.createElement('link');
          link.href = href;
          link.rel = rel;
          link.id = name;
          Object.entries(rest).forEach(([key, value]) => {
              link.setAttribute(key, value);
          });
          this.documentRef.head.appendChild(link);
          return new Promise((res, rej) => {
              link.onload = () => {
                  this.libraries.push(link);
                  res();
              };
              link.onerror = () => {
                  console.log('error linking lib');
                  rej();
              };
          });
      }
      logError(node, errorMessage) {
          this.emit('node:error', { node, errorMessage });
          console.error(errorMessage);
      }
      registerIframe() {
          this.iframe = this.rootDocumentRef.createElement('iframe');
          this.iframe.classList.add('page-iframe');
          this.container.appendChild(this.iframe);
      }
      registerStyleSheet() {
          const styleSheet = this.documentRef.createElement('style');
          styleSheet.setAttribute('type', 'text/css');
          this.documentRef.head.appendChild(styleSheet);
          this.styleSheet = styleSheet.sheet;
      }
      getStubElement() {
          return this.documentRef.createElement('span');
      }
  }

  function pageFactory(rootDocumentRef, container, title, schema, options) {
      return new Page(container, title, schema, options, rootDocumentRef);
  }

  exports.camelToKebabCase = camelToKebabCase;
  exports.hexToRgb = hexToRgb;
  exports.isCssPropNameValid = isCssPropNameValid;
  exports.isCssPropValueValid = isCssPropValueValid;
  exports.kebabToCamelCase = kebabToCamelCase;
  exports.pageFactory = pageFactory;
  exports.parseHTMLFragment = parseHTMLFragment;

}));
//# sourceMappingURL=index.js.map
