(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.HtmlPageBuilderStyles = {}));
})(this, (function (exports) { 'use strict';

  exports.StyleType = void 0;
  (function (StyleType) {
      StyleType["border"] = "border";
      StyleType["layout"] = "layout";
      StyleType["text"] = "text";
      StyleType["boxShadow"] = "boxShadow";
      StyleType["backgroundBlur"] = "backgroundBlur";
      StyleType["layerBlur"] = "layerBlur";
      StyleType["custom"] = "custom";
      StyleType["size"] = "size";
  })(exports.StyleType || (exports.StyleType = {}));

  // ChatGPT
  function camelToKebabCase(str) {
      return str.replace(/([A-Z])/g, '-$1').toLowerCase();
  }
  function hexToRgb(hex) {
      // Remove the '#' if it's included in the hex value
      hex = hex.replace(/^#/, '');
      // If the hex is shorthand (e.g. #03F), expand it to full form (e.g. #0033FF)
      if (hex.length === 3) {
          hex = hex.split('').map(char => char + char).join('');
      }
      // Parse the r, g, b values from the hex string
      const bigint = parseInt(hex, 16);
      const r = (bigint >> 16) & 255;
      const g = (bigint >> 8) & 255;
      const b = bigint & 255;
      return `rgb(${r}, ${g}, ${b})`;
  }
  function isHexColor(str) {
      // Regular expression to match valid hex color codes (#RRGGBB or #RGB)
      const hexRegex = /^#([A-Fa-f0-9]{3}){1,2}$/;
      return hexRegex.test(str);
  }
  function isCssPropValueValid(propertyName, propertyValue) {
      const propToCheck = camelToKebabCase(propertyName);
      const propValueToCheck = propertyValue.split(' ').map(item => isHexColor(item) ? hexToRgb(item) : item).join(' ');
      if (CSS.supports(propToCheck, propValueToCheck)) {
          return true;
      }
      // Optionally check for vendor prefixes
      const prefixes = ['-webkit-', '-moz-', '-ms-', '-o-'];
      return prefixes.some(prefix => CSS.supports(prefix + propToCheck, propertyValue));
  }

  class StyleRule {
      constructor() {
          this.rule = {};
          this.detachedProps = new Set();
      }
      getAllProps() {
          return Object.keys(this.rule).filter(key => !!this.rule[key]).reduce((r, key) => {
              r[key] = this.rule[key];
              return r;
          }, {});
      }
      setProp(key, value) {
          this.rule[key] = value;
      }
      getProp(key) {
          return this.rule[key];
      }
      getPropOrPropEdges(key, options) {
          return this.rule[key] ?? this.getPropEdges(key, options);
      }
      getPropEdges(key, options) {
          const edges = options?.edgeKeys ?? ['left', 'top', 'right', 'bottom'];
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          const edgesRule = edges.reduce((r, edge) => {
              const edgePropKey = edgePropKeyFactory(edge);
              if (this.rule[edgePropKey]) {
                  r[edge] = this.rule[edgePropKey];
              }
              return r;
          }, {});
          return Object.keys(edgesRule).length ? edgesRule : undefined;
      }
      setPropOrPropEdges(key, value, options) {
          if (typeof value === 'string') {
              this.deletePropEdges(key, options);
              this.rule[key] = value;
              return;
          }
          this.deleteProp(key);
          this.setPropEdges(key, value, options);
      }
      setPropEdges(key, value, options) {
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          Object.entries(value).forEach(([edgeKey, value]) => {
              const edgePropKey = edgePropKeyFactory(edgeKey);
              this.rule[edgePropKey] = value;
          });
      }
      attachProp(key) {
          this.detachedProps.delete(key);
      }
      detachProp(key) {
          this.detachedProps.add(key);
      }
      detached(key) {
          return this.detachedProps.has(key);
      }
      init(rule) {
          this.rule = rule;
      }
      deleteProp(key) {
          delete this.rule[key];
      }
      deletePropEdges(key, options) {
          const edges = options?.edgeKeys ?? ['left', 'top', 'right', 'bottom'];
          const edgePropKeyFactory = options?.getEdgePropKey ?? ((edgeKey) => `${key}-${edgeKey}`);
          edges.forEach(edgeKey => {
              const edgePropKey = edgePropKeyFactory(edgeKey);
              delete this.rule[edgePropKey];
          });
      }
      entries() {
          return Object.entries(this.rule);
      }
      generateRuleText() {
          return this.entries()
              .filter(([key, value]) => !this.detachedProps.has(key) && !!value)
              .reduce((result, [key, value]) => {
              const ruleText = `${key}: ${value};`;
              result += ruleText;
              return result;
          }, '');
      }
  }

  class StyleBase {
      constructor() {
          this.rule = new StyleRule();
          this.ruleIndex = -1;
      }
      setStyleRule(ruleValue) {
          this.rule.init(ruleValue);
          this.applyStyleRule();
      }
      applyStyles() {
          this.applyStyleRule();
          this.element.page.schemaChanged();
      }
      tryToApplyStyleRule(apply = true) {
          if (apply) {
              this.applyStyles();
          }
      }
      applyStyleRule() {
          const ruleText = this.rule.generateRuleText();
          if (!ruleText?.length && this.ruleIndex >= 0) {
              this.hardDeleteRule();
              this.updateNodeSchema();
              this.element.emit('view:change');
              return;
          }
          if (ruleText === this.ruleText) {
              return;
          }
          this.ruleText = ruleText;
          if (this.ruleIndex >= 0) {
              this.updateRule();
          }
          else {
              this.insertRule();
          }
          if (!this.element.nativeElement.classList.contains(this.selector)) {
              this.element.nativeElement.classList.add(this.selector);
          }
          this.updateNodeSchema();
          this.element.emit('view:change');
      }
      insertRule() {
          this.ruleIndex = this.styleSheet.insertRule(`.${this.selector}{${this.ruleText}}`, this.ruleIndex >= 0 ? this.ruleIndex : this.styleSheet.cssRules.length);
      }
      updateRule() {
          this.deleteRule();
          this.insertRule();
      }
      deleteRule() {
          this.styleSheet.deleteRule(this.ruleIndex);
      }
      hardDeleteRule() {
          if (!this.styleSheet.cssRules.length) {
              console.warn('No rules were applied, nothing to delete');
              return;
          }
          this.deleteRule();
          this.ruleText = '';
          this.ruleIndex = -1;
      }
      updateNodeSchema() {
          if (!this.ruleText?.length) {
              delete this.element.schema.styles?.[this.styleType];
          }
          else {
              this.element.schema.styles = {
                  ...this.element.schema.styles,
                  [this.styleType]: this.rule.getAllProps()
              };
          }
      }
  }

  class BackgroundBlurStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.backgroundBlur;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      set(blur, apply = true) {
          this.rule.setProp('backdrop-filter', `blur(${blur})`);
          this.tryToApplyStyleRule(apply);
      }
      get() {
          return this.getEffectValue();
      }
      delete(apply = true) {
          this.rule.deleteProp('backdrop-filter');
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(value) {
          return isCssPropValueValid('backdrop-filter', `blur(${value})`);
      }
      getEffectValue() {
          if (!this.rule.getProp('backdrop-filter')) {
              return '';
          }
          const match = this.rule.getProp('backdrop-filter').match(/\(([^)]+)\)/);
          return match[1];
      }
  }

  class BorderStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.border;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setWidth(width, apply = true) {
          this.setInitialValuesOnChange('width');
          if (typeof width !== 'string') {
              this.rule.setProp('border-width', '0px');
              this.rule.setPropEdges('border-width', width, {
                  getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
              });
          }
          else {
              this.rule.setProp('border-width', width);
              this.rule.deletePropEdges('border-width', {
                  getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
              });
          }
          this.tryToApplyStyleRule(apply);
      }
      getWidth() {
          return this.rule.getPropEdges('border-width', {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-width`
          }) ?? this.rule.getProp('border-width');
      }
      setStyleType(styleType, apply = true) {
          this.setInitialValuesOnChange('styleType');
          this.rule.setPropOrPropEdges('border-style', styleType, {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-style`
          });
          this.tryToApplyStyleRule(apply);
      }
      getStyleType() {
          return this.rule.getPropOrPropEdges('border-style');
      }
      setColor(color, apply = true) {
          this.setInitialValuesOnChange('color');
          this.rule.setPropOrPropEdges('border-color', color, {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-color`
          });
          this.tryToApplyStyleRule(apply);
      }
      getColor() {
          return this.rule.getPropOrPropEdges('border-color', {
              getEdgePropKey: (edgeKey) => `border-${edgeKey}-color`,
          });
      }
      setBorderRadius(radius, apply = true) {
          this.setInitialValuesOnChange('border-radius');
          this.rule.setPropOrPropEdges('border-radius', radius, {
              edgeKeys: ['topRight', 'topLeft', 'bottomRight', 'bottomLeft'],
              getEdgePropKey: (edgeKey) => `border-${camelToKebabCase(edgeKey)}-radius`
          });
          this.tryToApplyStyleRule(apply);
      }
      getBorderRadius() {
          return this.rule.getPropOrPropEdges('border-radius', {
              edgeKeys: ['topRight', 'topLeft', 'bottomRight', 'bottomLeft'],
              getEdgePropKey: (edgeKey) => `border-${camelToKebabCase(edgeKey)}-radius`
          });
      }
      setInitialValuesOnChange(changedProp) {
          if (changedProp !== 'width' && this.getWidth() === undefined) {
              this.rule.setProp('border-width', '1px');
          }
          if (changedProp !== 'styleType' && this.getStyleType() === undefined) {
              this.rule.setProp('border-style', 'solid');
          }
          if (changedProp !== 'color' && this.getColor() === undefined) {
              this.rule.setProp('border-color', '#000000');
          }
      }
  }

  class BoxShadowStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.boxShadow;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      has(effect) {
          const str = this.effectToString(effect);
          return !!this.rule.getProp('box-shadow')?.includes(str);
      }
      set(effect, apply = true) {
          const shadowProp = this.effectToString(effect);
          const boxShadowValue = this.rule.getProp('box-shadow') ?? '';
          const updatedBoxShadowValue = `${boxShadowValue}${boxShadowValue?.length ? ', ' : ''}${shadowProp}`;
          this.rule.setProp('box-shadow', updatedBoxShadowValue);
          this.tryToApplyStyleRule(apply);
      }
      getEffects() {
          return this.rule.getProp('box-shadow')?.split(', ').map((prop) => this.stringToEffect(prop)) || [];
      }
      delete(effect, apply = true) {
          if (!this.rule.getProp('box-shadow')?.length) {
              return;
          }
          const shadowProp = this.effectToString(effect);
          const boxShadowProps = this.rule.getProp('box-shadow').split(', ');
          this.rule.setProp('box-shadow', boxShadowProps.filter((prop) => shadowProp !== prop).join(', '));
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(effect) {
          return isCssPropValueValid('box-shadow', this.effectToString(effect));
      }
      effectToString(shadow) {
          return `${shadow.inner ? 'inset ' : ''}${shadow.x} ${shadow.y} ${shadow.blur} ${shadow.spread} ${shadow.color}`;
      }
      stringToEffect(str) {
          const items = str.split(' ');
          const inner = items[0] === 'inset';
          const [x, y, blur, spread, color] = items.filter(item => item !== 'inset');
          return { inner, x, y, blur, spread, color };
      }
  }

  class CustomStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.custom;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      get(property) {
          return this.rule.getProp(property);
      }
      getAll() {
          return this.rule.getAllProps();
      }
      append(property, value, apply = true) {
          this.rule.setProp(property, value);
          this.tryToApplyStyleRule(apply);
      }
      delete(property, apply = true) {
          this.rule.deleteProp(property);
          this.tryToApplyStyleRule(apply);
      }
      attach(key) {
          this.rule.attachProp(key);
          super.applyStyles();
      }
      detach(key) {
          this.rule.detachProp(key);
          super.applyStyles();
      }
      detached(key) {
          return this.rule.detached(key);
      }
  }

  class LayerBlurStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.layerBlur;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      set(blur, apply = true) {
          this.rule.setProp('filter', `blur(${blur})`);
          this.tryToApplyStyleRule(apply);
      }
      get() {
          return this.getEffectValue();
      }
      delete(apply = true) {
          this.rule.deleteProp('filter');
          this.tryToApplyStyleRule(apply);
      }
      isValidValue(value) {
          return isCssPropValueValid('filter', `blur(${value})`);
      }
      getEffectValue() {
          if (!this.rule.getProp('filter')) {
              return '';
          }
          const match = this.rule.getProp('filter').match(/\(([^)]+)\)/);
          return match[1];
      }
  }

  class LayoutStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.layout;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setPosition({ direction, x, y, gap }, apply = true) {
          this.rule.setProp('display', 'flex');
          this.rule.setProp('flex-direction', direction);
          this.rule.setProp('gap', gap ?? '');
          if (direction === 'row') {
              this.rule.setProp('justify-content', x === 'space-between' || x === 'center' ? x : `flex-${x}`);
              this.rule.setProp('align-items', y === 'center' ? y : `flex-${y}`);
          }
          else if (direction === 'column') {
              this.rule.setProp('justify-content', y === 'center' || y === 'space-between' ? y : `flex-${y}`);
              this.rule.setProp('align-items', x === 'center' ? x : `flex-${x}`);
          }
          this.tryToApplyStyleRule(apply);
      }
      getPosition() {
          if (this.rule.getProp('display') !== 'flex') {
              return undefined;
          }
          const p = {};
          p.direction = this.rule.getProp('flex-direction');
          let xRule, yRule;
          if (p.direction === 'row') {
              xRule = this.rule.getProp('justify-content');
              yRule = this.rule.getProp('align-items');
          }
          else {
              xRule = this.rule.getProp('align-items');
              yRule = this.rule.getProp('justify-content');
          }
          p.x = xRule === 'space-between' || xRule === 'center' ? xRule : xRule.split('-')[1];
          p.y = yRule === 'space-between' || yRule === 'center' ? yRule : yRule.split('-')[1];
          p.gap = this.rule.getProp('gap');
          return p;
      }
      deletePosition(apply = true) {
          this.rule.deleteProp('display');
          this.rule.deleteProp('flex-direction');
          this.rule.deleteProp('justify-content');
          this.rule.deleteProp('align-items');
          this.rule.deleteProp('gap');
          this.tryToApplyStyleRule(apply);
      }
      setPadding(padding, apply = true) {
          this.rule.setPropOrPropEdges('padding', padding);
          this.tryToApplyStyleRule(apply);
      }
      getPadding() {
          return this.rule.getPropOrPropEdges('padding');
      }
      deletePadding(apply = true) {
          this.rule.deleteProp('padding');
          this.rule.deletePropEdges('padding');
          this.tryToApplyStyleRule(apply);
      }
      setMargin(margin, apply = true) {
          this.rule.setPropOrPropEdges('margin', margin);
          this.tryToApplyStyleRule(apply);
      }
      getMargin() {
          return this.rule.getPropOrPropEdges('margin');
      }
      deleteMargin(apply = true) {
          this.rule.deleteProp('margin');
          this.tryToApplyStyleRule(apply);
      }
      setBackgroundColor(color, apply = true) {
          this.rule.setProp('background-color', color);
          this.tryToApplyStyleRule(apply);
      }
      getBackgroundColor() {
          return this.rule.getProp('background-color');
      }
      deleteBackgroundColor(apply = true) {
          this.rule.deleteProp('background-color');
          this.tryToApplyStyleRule(apply);
      }
  }

  class SizeStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.size;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
      }
      setHeight(height, apply = true) {
          this.rule.setProp('height', height);
          this.tryToApplyStyleRule(apply);
      }
      getHeight() {
          return this.rule.getProp('height');
      }
      deleteHeight(apply = true) {
          this.rule.deleteProp('height');
          this.tryToApplyStyleRule(apply);
      }
      setWidth(width, apply = true) {
          this.rule.setProp('width', width);
          this.tryToApplyStyleRule(apply);
      }
      getWidth() {
          return this.rule.getProp('width');
      }
      deleteWidth(apply = true) {
          this.rule.deleteProp('width');
          this.tryToApplyStyleRule(apply);
      }
  }

  class TextStyle extends StyleBase {
      constructor(element) {
          super();
          this.element = element;
          this.styleType = exports.StyleType.text;
          this.selector = `${this.styleType}-${this.element.key}`;
          this.styleSheet = this.element.page.getStyleSheet();
          this.setDefaults();
      }
      getColor() {
          return this.rule.getProp('color');
      }
      setColor(color, apply = true) {
          this.rule.setProp('color', color);
          this.tryToApplyStyleRule(apply);
      }
      getLineHeight() {
          return this.rule.getProp('line-height');
      }
      setLineHeight(lineHeight, apply = true) {
          this.rule.setProp('line-height', lineHeight);
          this.tryToApplyStyleRule(apply);
      }
      getAlignment() {
          return this.rule.getProp('text-align');
      }
      setAlignment(alignment, apply = true) {
          this.rule.setProp('text-align', alignment);
          this.tryToApplyStyleRule(apply);
      }
      getFontSize() {
          return this.rule.getProp('font-size');
      }
      setFontSize(fontSize, apply = true) {
          this.rule.setProp('font-size', fontSize);
          this.tryToApplyStyleRule(apply);
      }
      getDecoration() {
          return this.rule.getProp('text-decoration');
      }
      setDecoration(decoration, apply = true) {
          this.rule.setProp('text-decoration', decoration);
          this.tryToApplyStyleRule(apply);
      }
      setDefaults() {
          this.rule.setProp('color', '#000000');
          this.rule.setProp('line-height', 'auto');
          this.rule.setProp('text-align', 'left');
      }
  }

  exports.InvalidCssPropValue = void 0;
  (function (InvalidCssPropValue) {
      InvalidCssPropValue[InvalidCssPropValue["invalidProperty"] = 1] = "invalidProperty";
      InvalidCssPropValue[InvalidCssPropValue["invalidValue"] = 2] = "invalidValue";
  })(exports.InvalidCssPropValue || (exports.InvalidCssPropValue = {}));

  exports.BackgroundBlurStyle = BackgroundBlurStyle;
  exports.BorderStyle = BorderStyle;
  exports.BoxShadowStyle = BoxShadowStyle;
  exports.CustomStyle = CustomStyle;
  exports.LayerBlurStyle = LayerBlurStyle;
  exports.LayoutStyle = LayoutStyle;
  exports.SizeStyle = SizeStyle;
  exports.TextStyle = TextStyle;

}));
//# sourceMappingURL=index.js.map
